/*
 * input.c - Read references from input trace files. Supports reading directly
 *           from gzip files if zlib is available.
 *
 */

#define DEBUG_INPUT 0

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_ZLIB_H
#include <zlib.h>
#endif

#include <vmtrace.h>
#include <input.h>
#include <options.h>
#include <etrace.h>
#include <util.h>

#ifdef HAVE_LIBZ
typedef gzFile input_file_t;
static int _zerrno;
#define FDOPEN(fd, mode) gzdopen(fd, mode)
#define READ(f, buf, len) gzread(f, buf, len)
#define GETC(f) gzgetc(f)
#define FERROR(f, msg) fprintf(stderr, "%s: %s", msg, gzerror(f, &_zerrno))
#define FEOF(f) gzeof(f)
#else
typedef FILE * input_file_t;
#define FDOPEN(fd, mode) fdopen(fd, mode)
#define READ(f, buf, len) fread(buf, 1, len, f)
#define GETC(f) fgetc(f)
#define FERROR(f, msg) perror(msg)
#define FEOF(f) feof(f)
#endif

#define FD_STDIN 0

typedef struct _input_state {
  input_file_t f;
  vaddr_t nextinst, nextload, nextstore;
  int input_fd;
  int input_len; /* length of input, in bytes, if known (-1 if unknown) */
} input_state_t;

input_state_t g_in;

static input_file_t input_open(char *input_file, int *fd, int *input_len);
static int input_get_file_len(int fd);
static int input_fd_pos(int fd);

static inline byte_t input_read_byte(input_file_t f);
static uint input_read_dword(input_file_t f);
static inline bool_t input_eof(input_file_t f);

/* Routines to advance past constructs that we ignore */
static void input_consume_string(input_file_t f);
static void input_consume_module(input_state_t *in, byte_t key);
static void input_consume_thread(input_state_t *in, byte_t key);
static void input_consume_branch(input_state_t *in, byte_t key);
static void input_consume_proc(input_state_t *in, byte_t key);

void input_init(char *input_file) {
  g_in.f = input_open(input_file, &g_in.input_fd, &(g_in.input_len));
  g_in.nextinst = g_in.nextload = g_in.nextstore = 0;
}

input_file_t input_open(char *input_file, int *fd, int *input_len) {
  input_file_t stream;
  if (input_file == NULL) {
    *fd = FD_STDIN;
    *input_len = -1;
  } else {
    *fd = open(input_file, O_RDONLY, 0);
    if (*fd == -1) {
      perror("vmtrace: failed to open input file");
      abort();
    }
  }
  *input_len = input_get_file_len(*fd);
  stream = FDOPEN(*fd, "r");
  assert(stream);
  return stream;
}

/* Get the length of the file represented by fd.
 * Returns -1 if can't be determined.
 * This should work even if the shell is redirecting input from a file,
 * though not from another process (i.e. pipe). */
int input_get_file_len(int fd) {
  struct stat buf;
  if (fstat(fd, &buf) == -1) {
    perror("vmtrace: calculating file size");
    return -1;
  }
  /* stdin from another process will be reported as 0 bytes long */
  if (buf.st_size == 0)
    return -1;
  
  return buf.st_size;
}

int input_fd_pos(int fd) {
  return lseek(fd, 0L, SEEK_CUR);
}

void input_get_stats(int *curr, int *len) {
  *curr = input_fd_pos(g_in.input_fd);
  *len = g_in.input_len;
}

inline bool_t input_eof(input_file_t f) {
  return FEOF(f);
}

inline byte_t input_read_byte(input_file_t f) {
  return GETC(f);
}

uint input_read_dword(input_file_t f) {
  uint dword;
  int bytes;
  bytes = READ(f, &dword, sizeof(dword));
  if (bytes < 0) {
    FERROR(f, "vmtrace: read error");
    abort();
  }
  /* XXX */
  if (bytes < sizeof(dword)) {
    fprintf(stderr, "incomplete dword read\n");
    abort();
  }
  return dword;
}

/* Eat up a string from the file. Not the most efficient, but
 * hopefully not needed too often (only on module load). */
void input_consume_string(input_file_t f) {
  while (input_read_byte(f) && !input_eof(f)) {
    ;
  }  
}

void input_consume_module(input_state_t *in, byte_t key) {
  /* mid: */
  input_read_byte(in->f);
  if (GET_DESC(key) == MT_LOAD) {
    /* nprocs: */
    input_read_dword(in->f);
    /* nbranches: */
    input_read_dword(in->f);
    /* base_address: */
    input_read_dword(in->f);
    /* module name: */
    input_consume_string(in->f);
  }
}

void input_consume_thread(input_state_t *in, byte_t key) {
  /* tid: */
  input_read_byte(in->f);
}

void input_consume_branch(input_state_t *in, byte_t key) {
  /* bid: */
  input_read_dword(in->f);
  /* target: */
  input_read_dword(in->f);
}

void input_consume_proc(input_state_t *in, byte_t key) {
//  printf("consume proc\n");
  /* pid: */
  input_read_dword(in->f);
  /* esp: */
  input_read_dword(in->f);
}

bool_t input_next_ref(vaddr_t *addr, ref_kind_t *type, uint *length) {
  input_state_t *in = &g_in;
  byte_t key;
  bool_t is_next = TRUE;
  vaddr_t *next = NULL;

  do {
    key = input_read_byte(in->f);
    if (input_eof(in->f) || key == 0xFF || key == ET_EOF) return FALSE;
    
    switch (ET_TYPE(key)) {
    case ET_MODULE:
      input_consume_module(in, key);
      break;
    case ET_THREAD:
      input_consume_thread(in, key);
      break;
    case ET_BRANCH:
      input_consume_branch(in, key);
      break;
    case ET_PROC:
      input_consume_proc(in, key);
      break;
      
    case ET_INST:
      is_next = FALSE;
    case ET_INSTNEXT:
      next = &(in->nextinst);
      *type = REF_KIND_CODE;
      break;
      
    case ET_LOAD:
      is_next = FALSE;
    case ET_LOADNEXT:
      next = &(in->nextload);
      *type = REF_KIND_LOAD;
      break;
      
    case ET_STORE:
      is_next = FALSE;
    case ET_STORENEXT:
      next = &(in->nextstore);
      *type = REF_KIND_STORE;
      break;

    case ET_EOF:
      printf("hit ET_EOF\n");
      return FALSE;
      
    default:
      printf("ignoring key 0x%x\n", key);
    }
  } while (next == NULL);

  *addr = (!is_next) ? file_to_host_uint(input_read_dword(in->f)) : *next;
  *next = *addr + ET_LENGTH(key);
  *length = ET_LENGTH(key);

#if DEBUG_INPUT
  printf("addr: %x, length: %x, type: %d, is_next: %d\n", *addr, *length, *type, is_next);
#endif
  return TRUE;
}

